# sage_setup: distribution = sagemath-categories
"""nodoctest
all.py -- export of projective schemes to Sage
"""

# ****************************************************************************
#  Copyright (C) 2005 William Stein <wstein@gmail.com>
#  Copyright (C) 2025 Gareth Ma <grhkm21@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.schemes.weighted_projective.weighted_projective_space import WeightedProjectiveSpace



# sage_setup: distribution = sagemath-categories
"""
Hom-sets of weighted projective schemes

AUTHORS:

- Gareth Ma (2024): initial version, based on unweighted version.
"""

# *****************************************************************************
#        Copyright (C) 2006 William Stein <wstein@gmail.com>
#        Copyright (C) 2011 Volker Braun <vbraun.name@gmail.com>
#        Copyright (C) 2024 Gareth Ma <grhkm21@gmail.com>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#   as published by the Free Software Foundation; either version 2 of
#   the License, or (at your option) any later version.
#                   http://www.gnu.org/licenses/
# *****************************************************************************

from sage.schemes.generic.homset import SchemeHomset_points


class SchemeHomset_points_weighted_projective_ring(SchemeHomset_points):
    """
    Set of rational points of a weighted projective variety over a ring.

    INPUT:

    See :class:`SchemeHomset_points`.

    EXAMPLES::

        sage: W = WeightedProjectiveSpace([3, 4, 5], QQ)
        sage: W.point_homset()
        Set of rational points of Weighted Projective Space of dimension 2 with weights (3, 4, 5) over Rational Field
        sage: W.an_element().parent() is W.point_homset()
        True
    """



# sage_setup: distribution = sagemath-categories
from sage.categories.fields import Fields
from sage.categories.map import Map
from sage.misc.latex import latex
from sage.misc.prandom import shuffle
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.multi_polynomial_ring_base import MPolynomialRing_base
from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.polynomial.term_order import TermOrder
from sage.schemes.generic.ambient_space import AmbientSpace
from sage.schemes.projective.projective_space import ProjectiveSpace, _CommRings
from sage.structure.all import UniqueRepresentation
from sage.structure.category_object import normalize_names

from sage.schemes.weighted_projective.weighted_projective_homset import SchemeHomset_points_weighted_projective_ring


def WeightedProjectiveSpace(weights, R=None, names=None):
    r"""
    Return a weighted projective space with the given ``weights`` over the ring ``R``.

    EXAMPLES::

        sage: WP = WeightedProjectiveSpace([1, 3, 1]); WP
        Weighted Projective Space of dimension 2 with weights (1, 3, 1) over Integer Ring
    """
    if isinstance(weights, (MPolynomialRing_base, PolynomialRing_generic)) and R is None:
        if names is not None:
            # Check for the case that the user provided a variable name
            # That does not match what we wanted to use from R
            names = normalize_names(weights.ngens(), names)
            if weights.variable_names() != names:
                # The provided name doesn't match the name of R's variables
                raise NameError(
                    "variable names passed to ProjectiveSpace conflict with names in ring"
                )
        A = WeightedProjectiveSpace(
            weights.ngens() - 1, weights.base_ring(), names=weights.variable_names()
        )
        A._coordinate_ring = weights
        return A

    if isinstance(R, (int, Integer, list, tuple)):
        weights, R = R, weights
    elif R is None:
        R = ZZ

    # WeightedProjectiveSpace(5) -> just return unweighted version
    if isinstance(weights, (int, Integer)):
        return ProjectiveSpace(weights, R=R, names=names)
    elif isinstance(weights, (list, tuple)):
        # Make it hashable
        weights = tuple(map(Integer, weights))
        if any(w <= 0 for w in weights):
            raise TypeError(
                f"weights(={weights}) should only consist of positive integers"
            )
    else:
        raise TypeError(f"weights={weights} must be an integer, a list or a tuple")

    if names is None:
        names = "x"

    # TODO: Specialise implementation to projective spaces over non-rings.
    if R in _CommRings:
        return WeightedProjectiveSpace_ring(weights, R=R, names=names)

    raise TypeError(f"R (={R}) must be a commutative ring")


class WeightedProjectiveSpace_ring(UniqueRepresentation, AmbientSpace):
    """
    Weighted projective space with the given ``weights`` over the ring `R`.

    EXAMPLES::

        sage: WeightedProjectiveSpace(Zp(5), [1, 3, 1], 'y')                        # needs sage.rings.padics
        Weighted Projective Space of dimension 2 with weights (1, 3, 1) over 5-adic Ring with
        capped relative precision 20
        sage: WeightedProjectiveSpace(QQ, 5, 'y')
        Projective Space of dimension 5 over Rational Field
        sage: _ is ProjectiveSpace(QQ, 5, 'y')
        True
    """

    @staticmethod
    def __classcall__(cls, weights: tuple[Integer], R=ZZ, names=None):
        # __classcall_ is the "preprocessing" step for UniqueRepresentation
        # see docs of CachedRepresentation
        # weights should be a tuple, also because it should be hashable
        if not isinstance(weights, tuple):
            raise TypeError(
                f"weights(={weights}) is not a tuple. Please use the"
                "`WeightedProjectiveSpace` constructor"
            )

        normalized_names = normalize_names(len(weights), names)
        return super().__classcall__(cls, weights, R, normalized_names)

    def __init__(self, weights: tuple[Integer], R=ZZ, names=None):
        """
        Initialization function.
        """
        AmbientSpace.__init__(self, len(weights) - 1, R)
        self._weights = weights
        self._assign_names(names)

    def weights(self) -> tuple[Integer]:
        """
        Return the tuple of weights of this weighted projective space.

        EXAMPLES::

            sage: WeightedProjectiveSpace(QQ, [1, 3, 1]).weights()
            (1, 3, 1)
        """
        return self._weights

    def ngens(self) -> Integer:
        """
        Return the number of generators of this weighted projective space.

        This is the number of variables in the coordinate ring of ``self``.

        EXAMPLES::

            sage: WeightedProjectiveSpace(QQ, [1, 3, 1]).ngens()
            3
            sage: WeightedProjectiveSpace(ZZ, 5).ngens()
            6
        """
        return self.dimension_relative() + 1

    def _check_satisfies_equations(self, v: list[Integer] | tuple[Integer]) -> bool:
        """
        Return ``True`` if ``v`` defines a point on the weighted projective
        plane; raise a :class:`TypeError` otherwise.

        EXAMPLES::

            sage: P = WeightedProjectiveSpace(ZZ, [1, 3, 1])
            sage: P._check_satisfies_equations([1, 1, 0])
            True

            sage: P._check_satisfies_equations([1, 0])
            Traceback (most recent call last):
            ...
            TypeError: the list v=[1, 0] must have 3 components

            sage: P._check_satisfies_equations([1/2, 0, 1])
            Traceback (most recent call last):
            ...
            TypeError: no conversion of this rational to integer
        """
        if not isinstance(v, (list, tuple)):
            raise TypeError(f"the argument v={v} must be a list or tuple")

        n = self.ngens()
        if len(v) != n:
            raise TypeError(f"the list v={v} must have {n} components")

        v = list(map(self.base_ring(), v))
        if all(vi.is_zero() for vi in v):
            raise TypeError("the zero vector is not a point in projective space")

        return True

    def coordinate_ring(self) -> PolynomialRing_generic:
        """
        Return the coordinate ring of this weighted projective space.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: WP = WeightedProjectiveSpace(GF(19^2, 'α'), [1, 3, 4, 1], 'abcd')
            sage: # needs sage.rings.finite_rings
            sage: R = WP.coordinate_ring(); R
            Multivariate Polynomial Ring in a, b, c, d over Finite Field in α of size 19^2
            sage: R.term_order()
            Weighted degree reverse lexicographic term order with weights (1, 3, 4, 1)

        ::

            sage: WP = WeightedProjectiveSpace(QQ, [1, 1, 1], ['alpha', 'beta', 'gamma'])
            sage: R = WP.coordinate_ring(); R
            Multivariate Polynomial Ring in alpha, beta, gamma over Rational Field
            sage: R.term_order()
            Weighted degree reverse lexicographic term order with weights (1, 1, 1)
        """
        if not hasattr(self, "_coordinate_ring"):  # we don't use cached_method to allow override in WeightedProjectiveSpace
            term_order = TermOrder("wdegrevlex", self.weights())
            self._coordinate_ring = PolynomialRing(
                self.base_ring(),
                self.dimension_relative() + 1,
                names=self.variable_names(),
                order=term_order,
            )
        return self._coordinate_ring

    def _validate(self, polynomials):
        """
        If ``polynomials`` is a tuple of valid polynomial functions on ``self``,
        return ``polynomials``, otherwise raise ``TypeError``.

        Since this is a weighted projective space, polynomials must be
        homogeneous with respect to the grading of this space.

        INPUT:

        - ``polynomials`` -- tuple of polynomials in the coordinate ring of
            this space.

        OUTPUT:

        - tuple of polynomials in the coordinate ring of this space.

        EXAMPLES::

            sage: P.<x, y, z> = WeightedProjectiveSpace(QQ, [1, 3, 1])
            sage: P._validate([x*y - z^4, x])
            [x*y - z^4, x]
            sage: P._validate([x*y - z^2, x])
            Traceback (most recent call last):
            ...
            TypeError: x*y - z^2 is not homogeneous with weights (1, 3, 1)
            sage: P._validate(x*y - z)
            Traceback (most recent call last):
            ...
            TypeError: the argument polynomials=x*y - z must be a list or tuple
        """
        if not isinstance(polynomials, (list, tuple)):
            raise TypeError(
                f"the argument polynomials={polynomials} must be a list or tuple"
            )

        R = self.coordinate_ring()
        for f in map(R, polynomials):
            if not f.is_homogeneous():
                raise TypeError(f"{f} is not homogeneous with weights {self.weights()}")

        return polynomials

    def _latex_(self) -> str:
        r"""
        Return a LaTeX representation of this weighted projective space.

        EXAMPLES::

            sage: print(latex(WeightedProjectiveSpace(ZZ, [1, 3, 1], 'x')))
            {\mathbf P}_{\Bold{Z}}^{[1, 3, 1]}

        TESTS::

            sage: WeightedProjectiveSpace(Zp(5), [2, 1, 3], 'y')._latex_()              # needs sage.rings.padics
            '{\\mathbf P}_{\\Bold{Z}_{5}}^{[2, 1, 3]}'
        """
        return (
            fr"{{\mathbf P}}_{{{latex(self.base_ring())}}}^{{{list(self.weights())}}}"
        )

    def _morphism(self, *_, **__):
        """
        Construct a morphism.

        For internal use only. See :mod:`morphism` for details.
        """
        raise NotImplementedError(
            "_morphism not implemented for weighted projective space"
        )

    def _homset(self, *_, **__):
        """
        Construct the Hom-set.
        """
        raise NotImplementedError(
            "_homset not implemented for weighted projective space"
        )

    def _point_homset(self, *args, **kwds):
        """
        Construct a point Hom-set.

        For internal use only. See :mod:`morphism` for details.
        """
        return SchemeHomset_points_weighted_projective_ring(*args, **kwds)

    def point(self, v, check: bool = True):
        """
        Create a point on this weighted projective space.

        INPUT:

        INPUT:

        - ``v`` -- anything that defines a point

        - ``check`` -- boolean (default: ``True``); whether
          to check the defining data for consistency

        OUTPUT: A point of this weighted projective space.

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace(QQ, [1, 3, 1])
            sage: WP.point([2, 3, 1])
            (2 : 3 : 1)
        """
        from sage.rings.infinity import infinity

        if v is infinity or (
            isinstance(v, (list, tuple)) and len(v) == 1 and v[0] is infinity
        ):
            if self.dimension_relative() > 1:
                raise ValueError("%s not well defined in dimension > 1" % v)
            v = [1, 0]

        return self.point_homset()(v, check=check)

    def _point(self, *args, **kwds):
        """
        Construct a point.

        For internal use only. See :mod:`morphism` for details.
        """
        from sage.schemes.weighted_projective.weighted_projective_point import SchemeMorphism_point_weighted_projective_ring
        return SchemeMorphism_point_weighted_projective_ring(*args, **kwds)

    def _repr_(self) -> str:
        """
        Return a string representation of this weighted projective space.

        EXAMPLES::

            sage: WeightedProjectiveSpace(Qp(5), [1, 3, 1], 'x')                        # needs sage.rings.padics
            Weighted Projective Space of dimension 2 with weights (1, 3, 1)
            over 5-adic Field with capped relative precision 20
        """
        return (
            f"Weighted Projective Space of dimension {self.dimension_relative()} with weights"
            f" {self.weights()} over {self.base_ring()}"
        )

    def change_ring(self, R):
        r"""
        Return a weighted projective space over ring ``R``.

        INPUT:

        - ``R`` -- commutative ring or morphism

        OUTPUT: weighted projective space over ``R``.
        If ``R`` is a morphism, return a weighted projective space over its codomain.

        .. NOTE::

            There is no need to have any relation between ``R`` and the base ring
            of this space, if you want to have such a relation, use
            ``self.base_extend(R)`` instead.

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([1, 3, 1], ZZ); WP
            Weighted Projective Space of dimension 2 with weights (1, 3, 1) over Integer Ring
            sage: WP.change_ring(QQ)
            Weighted Projective Space of dimension 2 with weights (1, 3, 1) over Rational Field
            sage: WP.change_ring(GF(5))
            Weighted Projective Space of dimension 2 with weights (1, 3, 1) over Finite Field of size 5
        """
        if isinstance(R, Map):
            return WeightedProjectiveSpace(self.weights(), R.codomain(),
                                           self.variable_names())
        else:
            return WeightedProjectiveSpace(self.weights(), R,
                                           self.variable_names())

    def _an_element_(self):
        r"""
        Return a (preferably typical) element of this space.

        This is used both for illustration and testing purposes.

        OUTPUT: a point in this weighted projective space.

        EXAMPLES::

            sage: WeightedProjectiveSpace(ZZ, [1, 3, 1], 'x').an_element()  # random
            (1 : 2 : 3)
            sage: WeightedProjectiveSpace(ZZ["y"], [2, 3, 1], 'x').an_element()  # random
            (3*y : 2*y : y)
        """
        n = self.dimension_relative()
        R = self.base_ring()
        coords = [(n + 1 - i) * R.an_element() for i in range(n + 1)]
        shuffle(coords)
        return self(coords)

    def subscheme(self, *_, **__):
        raise NotImplementedError("subscheme of weighted projective space has not been implemented")

    def curve(self, F):
        r"""
        Return a curve defined by ``F`` in this weighted projective space.

        INPUT:

        - ``F`` -- a polynomial, or a list or tuple of polynomials in
          the coordinate ring of this weighted projective space

        EXAMPLES::

            sage: WP.<x, y, z> = WeightedProjectiveSpace([1, 3, 1], QQ)
            sage: WP.curve(y^2 - x^5 * z - 3 * x^2 * z^4 - 2 * z^6)                     # needs sage.schemes
            Weighted Projective Curve over Rational Field defined by y^2 - x^5*z - 3*x^2*z^4 - 2*z^6
        """
        if self.base_ring() not in Fields():
            raise NotImplementedError("curves in weighted projective space over"
                                      "rings not implemented")
        from sage.schemes.curves.constructor import Curve
        return Curve(F, self)



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.graphs
r"""
Lattice posets
"""
# ****************************************************************************
#  Copyright (C) 2011 Nicolas M. Thiéry <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category import Category
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.posets import Posets
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import LazyImport


class LatticePosets(Category):
    r"""
    The category of lattices, i.e. partially ordered sets in which any
    two elements have a unique supremum (the elements' least upper
    bound; called their *join*) and a unique infimum (greatest lower bound;
    called their *meet*).

    EXAMPLES::

        sage: LatticePosets()
        Category of lattice posets
        sage: LatticePosets().super_categories()
        [Category of posets]
        sage: LatticePosets().example()
        NotImplemented

    .. SEEALSO::

        - :class:`~sage.categories.posets.Posets`
        - :class:`FiniteLatticePosets`, :func:`LatticePoset`

    TESTS::

        sage: C = LatticePosets()
        sage: TestSuite(C).run()
    """
    @cached_method
    def super_categories(self) -> list:
        r"""
        Return a list of the (immediate) super categories of
        ``self``, as per :meth:`Category.super_categories`.

        EXAMPLES::

            sage: LatticePosets().super_categories()
            [Category of posets]
        """
        return [Posets()]

    class ParentMethods:

        @abstract_method
        def meet(self, x, y):
            """
            Return the meet of `x` and `y` in this lattice.

            INPUT:

            - ``x``, ``y`` -- elements of ``self``

            EXAMPLES::

                sage: D = LatticePoset((divisors(30), attrcall("divides")))             # needs sage.graphs sage.modules
                sage: D.meet( D(6), D(15) )                                             # needs sage.graphs sage.modules
                3
            """

        @abstract_method
        def join(self, x, y):
            """
            Return the join of `x` and `y` in this lattice.

            INPUT:

            - ``x``, ``y`` -- elements of ``self``

            EXAMPLES::

                sage: D = LatticePoset((divisors(60), attrcall("divides")))             # needs sage.graphs sage.modules
                sage: D.join( D(6), D(10) )                                             # needs sage.graphs sage.modules
                30
            """

    class SubcategoryMethods:
        def ChainGraded(self):
            r"""
            A lattice is graded if all maximal chains have the same length.

            To avoid possible confusion, the name of the axiom
            is ``ChainGraded``.

            EXAMPLES::

                sage: P = posets.DivisorLattice(24)
                sage: P in FiniteLatticePosets().ChainGraded()
                True
            """
            return self._with_axiom("ChainGraded")

        def Stone(self):
            r"""
            A Stone lattice `(L, \vee, \wedge)` is a pseudo-complemented
            distributive lattice such that `a^* \vee a^{**} = 1`.

            See :wikipedia:`Stone algebra`.

            EXAMPLES::

                sage: P = posets.DivisorLattice(24)
                sage: P in FiniteLatticePosets().Stone()
                True
            """
            return self._with_axiom("Stone")

        def Distributive(self):
            r"""
            A lattice `(L, \vee, \wedge)` is distributive if meet
            distributes over join: `x \wedge (y \vee z) = (x \wedge y)
            \vee (x \wedge z)` for every `x,y,z \in L`.

            From duality in lattices, it follows that then also join
            distributes over meet.

            A distributive lattice is always graded.

            See :wikipedia:`Distributive lattice`.

            EXAMPLES::

                sage: # needs sage.modules
                sage: P = posets.ChainPoset(2).order_ideals_lattice()
                sage: P in FiniteLatticePosets().Distributive()
                True
            """
            return self._with_axiom("Trim")._with_axiom("ChainGraded")

        def CongruenceUniform(self):
            r"""
            A finite lattice `(L, \vee, \wedge)` is congruence uniform if it
            can be constructed by a sequence of interval doublings
            starting with the lattice with one element.

            EXAMPLES::

                sage: P = posets.TamariLattice(2)
                sage: P in FiniteLatticePosets().CongruenceUniform()
                True
            """
            return self._with_axiom("CongruenceUniform")

        def Semidistributive(self):
            r"""
            A finite lattice `(L, \vee, \wedge)` is semidistributive if
            it is both join-semidistributive and meet-semidistributive.

            A finite lattice is join-semidistributive if
            for all elements `e, x, y` in the lattice we have

            .. MATH::

                e \vee x = e \vee y \implies e \vee x = e \vee (x \wedge y)

            Meet-semidistributivity is the dual property.

            EXAMPLES::

                sage: P = posets.TamariLattice(2)
                sage: P in FiniteLatticePosets().Semidistributive()
                True
            """
            return self._with_axiom("Semidistributive")

        def Trim(self):
            r"""
            A finite lattice `(L, \vee, \wedge)` is trim if it is extremal
            and left modular.

            This notion is defined in [Thom2006]_.

            EXAMPLES::

                sage: P = posets.TamariLattice(2)
                sage: P in FiniteLatticePosets().Trim()
                True
            """
            return self._with_axiom("Trim")

        def Extremal(self):
            r"""
            A finite lattice `(L, \vee, \wedge)` is extremal if
            if it has a chain of length `n` (containing `n+1` elements)
            and exactly `n` join-irreducibles and `n` meet-irreducibles.

            This notion was defined by George Markowsky.

            EXAMPLES::

                sage: P = posets.TamariLattice(2)
                sage: P in FiniteLatticePosets().Extremal()
                True
            """
            return self._with_axiom("Extremal")

    Finite = LazyImport('sage.categories.finite_lattice_posets',
                        'FiniteLatticePosets')

    class Extremal(CategoryWithAxiom):
        """
        The category of extremal uniform lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().Extremal(); cat
            Category of finite extremal lattice posets

            sage: cat.super_categories()
            [Category of finite lattice posets,
             Category of extremal lattice posets]
        """
        class ParentMethods:
            def is_extremal(self) -> bool:
                """
                Return whether ``self`` is an extremal lattice.

                EXAMPLES::

                    sage: posets.TamariLattice(4).is_extremal()
                    True
                """
                return True

    class Trim(CategoryWithAxiom):
        """
        The category of trim uniform lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().Trim(); cat
            Category of finite trim lattice posets
            sage: cat.super_categories()
            [Category of finite lattice posets,
             Category of trim lattice posets]
        """
        @cached_method
        def extra_super_categories(self) -> list:
            r"""
            Return a list of the super categories of ``self``.

            This encode implications between properties.

            EXAMPLES::

                sage: FiniteLatticePosets().Trim().super_categories()
                [Category of finite lattice posets,
                 Category of trim lattice posets]
            """
            return [LatticePosets().Extremal()]

        class ParentMethods:
            def is_trim(self) -> bool:
                """
                Return whether ``self`` is a trim lattice.

                EXAMPLES::

                    sage: posets.TamariLattice(4).is_trim()                             # needs sage.modules
                    True
                """
                return True

    class Semidistributive(CategoryWithAxiom):
        """
        The category of semidistributive lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().Semidistributive(); cat
            Category of finite semidistributive lattice posets

            sage: cat.super_categories()
            [Category of finite lattice posets,
             Category of semidistributive lattice posets]
        """
        class ParentMethods:
            def is_semidistributive(self) -> bool:
                """
                Return whether ``self`` is a semidistributive lattice.

                EXAMPLES::

                    sage: posets.TamariLattice(4).is_semidistributive()
                    True
                """
                return True

    class CongruenceUniform(CategoryWithAxiom):
        """
        The category of congruence uniform lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().CongruenceUniform(); cat
            Category of finite congruence uniform lattice posets
            sage: cat.super_categories()
            [Category of finite lattice posets,
             Category of congruence uniform lattice posets]
        """
        @cached_method
        def extra_super_categories(self) -> list:
            r"""
            Return a list of the super categories of ``self``.

            This encode implications between properties.

            EXAMPLES::

                sage: FiniteLatticePosets().CongruenceUniform().super_categories()
                [Category of finite lattice posets,
                 Category of congruence uniform lattice posets]
            """
            return [LatticePosets().Semidistributive()]

        class ParentMethods:
            def is_congruence_uniform(self) -> bool:
                """
                Return whether ``self`` is a congruence uniform lattice.

                EXAMPLES::

                    sage: posets.TamariLattice(4).is_congruence_uniform()               # needs sage.modules
                    True
                """
                return True

    class Stone(CategoryWithAxiom):
        """
        The category of Stone lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().Stone(); cat
            Category of finite stone distributive lattices

            sage: cat.super_categories()
            [Category of finite distributive lattices,
             Category of stone lattice posets]
        """
        @cached_method
        def extra_super_categories(self) -> list:
            r"""
            Return a list of the super categories of ``self``.

            This encode implications between properties.

            EXAMPLES::

                sage: FiniteLatticePosets().Stone().super_categories()
                [Category of finite distributive lattices,
                 Category of stone lattice posets]
            """
            return [LatticePosets().Trim().ChainGraded()]

        class ParentMethods:
            def is_stone(self) -> bool:
                """
                Return whether ``self`` is a Stone lattice.

                EXAMPLES::

                    sage: posets.DivisorLattice(12).is_stone()                          # needs sage.modules
                    True
                """
                return True

    class ChainGraded(CategoryWithAxiom):
        """
        The category of graded lattices.

        EXAMPLES::

            sage: cat = FiniteLatticePosets().ChainGraded(); cat
            Category of finite chain graded lattice posets

            sage: cat.super_categories()
            [Category of finite lattice posets,
             Category of chain graded lattice posets]
        """
        class ParentMethods:
            def is_graded(self) -> bool:
                """
                Return whether ``self`` is a graded lattice.

                EXAMPLES::

                    sage: posets.DivisorLattice(12).is_graded()
                    True
                """
                return True


# the following was moved out of the main class

class DistributiveLattices(CategoryWithAxiom):
    """
    The category of distributive lattices.

    EXAMPLES::

        sage: cat = FiniteLatticePosets().Distributive(); cat
        Category of finite distributive lattices

        sage: cat.super_categories()
        [Category of finite lattice posets,
         Category of distributive lattices]

    TESTS::

        sage: from sage.categories.lattice_posets import DistributiveLattices
        sage: LatticePosets().Distributive() is DistributiveLattices()
        True
    """
    _base_category_class_and_axiom = (LatticePosets.Trim,
                                      "ChainGraded")

    class Finite(CategoryWithAxiom):
        pass

    class ParentMethods:
        def is_distributive(self) -> bool:
            """
            Return whether ``self`` is a distributive lattice.

            EXAMPLES::

                sage: # needs sage.modules
                sage: P = posets.Crown(4).order_ideals_lattice()
                sage: P.is_distributive()
                True
            """
            return True


LatticePosets.Trim.ChainGraded = DistributiveLattices



# sage_setup: distribution = sagemath-categories
# distutils: language = c++
# distutils: extra_compile_args = -std=c++17

# ******************************************************************************
# Copyright (C) 2024 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ******************************************************************************

from cpython cimport PyObject
from libcpp.pair cimport pair
from sage.data_structures.bitset_base cimport bitset_t


# ==============================================================================
# Interface to pairing heap data structure from ./pairing_heap.h
# ==============================================================================

cdef extern from "./pairing_heap.h" namespace "pairing_heap":
    cdef cppclass PairingHeap[TypeOfItem, TypeOfValue]:
        PairingHeap() except +
        PairingHeap(PairingHeap[TypeOfItem, TypeOfValue]) except +
        bint empty()
        void push(TypeOfItem, TypeOfValue) except +
        pair[TypeOfItem, TypeOfValue] top() except +
        TypeOfItem top_item() except +
        TypeOfValue top_value() except +
        void pop() except +
        void decrease(TypeOfItem, TypeOfValue) except +
        bint contains(TypeOfItem)
        TypeOfValue value(TypeOfItem) except +
        size_t size()

    cdef cppclass PairingHeapNodePy:
        PyObject * value           # value associated with the item
        PairingHeapNodePy * prev   # Previous sibling of the node or parent
        PairingHeapNodePy * next   # Next sibling of the node
        PairingHeapNodePy * child  # First child of the node

        @staticmethod
        PairingHeapNodePy * _merge(PairingHeapNodePy * a, PairingHeapNodePy * b) except +

        @staticmethod
        PairingHeapNodePy * _pair(PairingHeapNodePy * p) except +

        @staticmethod
        void _link(PairingHeapNodePy * a, PairingHeapNodePy * b)

        @staticmethod
        void _unlink(PairingHeapNodePy * p)


# ==============================================================================
# Pairing heap data structure with fixed capacity n
# ==============================================================================

cdef class PairingHeap_class:
    cdef size_t n                   # maximum number of items
    cdef PairingHeapNodePy * root   # pointer to the top of the heap
    cdef PairingHeapNodePy * nodes  # array of size n to store items
    cdef size_t number_of_items     # number of active items
    cpdef bint empty(self) noexcept
    cpdef bint full(self) noexcept
    cpdef size_t capacity(self) noexcept
    cpdef size_t size(self) noexcept
    cpdef tuple top(self)
    cpdef object top_value(self)
    cpdef void pop(self) noexcept


cdef class PairingHeap_of_n_integers(PairingHeap_class):
    cdef bitset_t active  # bitset to identify active items
    cpdef void push(self, size_t item, object value) except *
    cpdef size_t top_item(self) except *
    cpdef void decrease(self, size_t item, object new_value) except *
    cpdef object value(self, size_t item)


cdef class PairingHeap_of_n_hashables(PairingHeap_class):
    cdef list _int_to_item  # mapping from integers to items
    cdef dict _item_to_int  # mapping from items to integers
    cdef list free_idx      # list of free indexes
    cpdef void push(self, object item, object value) except *
    cpdef object top_item(self)
    cpdef void decrease(self, object item, object new_value) except *
    cpdef object value(self, object item)



# sage_setup: distribution = sagemath-objects
from cpython.ref import PyTypeObject
from sage.libs.gmp.types import mpz_ptr

def Integer_AS_MPZ(x: object) -> mpz_ptr:
    ...

def is_Integer(x: object) -> bool:
    ...



# sage_setup: distribution = sagemath-categories
from sage.libs.gmp.types import __mpz_struct, mpz_t, mpz_ptr
from sage.libs.gmp.mpz import mpz_set

from sage.structure.element import EuclideanDomainElement, RingElement
from sage.categories.morphism import Morphism

class Integer(EuclideanDomainElement):
    value: __mpz_struct[1]

    def set_from_mpz(self, value: mpz_t) -> None:
        ...

    def hash_c(self) -> int:
        ...

    def __pari__(self) -> object:
        ...

    def _shift_helper(self, y: object, sign: int) -> object:
        ...

    def _add_(self, other: object) -> object:
        ...

    def _mul_(self, other: object) -> object:
        ...

    def _pow_(self, other: object) -> object:
        ...

    def _and(self, other: 'Integer') -> 'Integer':
        ...

    def _or(self, other: 'Integer') -> 'Integer':
        ...

    def _xor(self, other: 'Integer') -> 'Integer':
        ...

    def _exact_log_log2_iter(self, m: 'Integer') -> int:
        ...

    def _exact_log_mpfi_log(self, m: object) -> int:
        ...

    def _valuation(self, p: 'Integer') -> RingElement:
        ...

    def _val_unit(self, p: 'Integer') -> object:
        ...

    def _divide_knowing_divisible_by(self, right: 'Integer') -> 'Integer':
        ...

    def _is_power_of(self, n: 'Integer') -> bool:
        ...

    def _pseudoprime_is_prime(self, proof: object) -> bool:
        ...

def mpz_set_str_python(z: mpz_ptr, s: str, base: int) -> int:
    ...

def smallInteger(value: int) -> 'Integer':
    ...

_small_primes_table: list[bool]

def _Integer_from_mpz(e: mpz_t) -> 'Integer':
    z = Integer.__new__(Integer)
    mpz_set(z.value, e)
    return z



# sage_setup: distribution = sagemath-categories
from typing import Any, Union

def is_FractionFieldElement(x: Any) -> bool:
    ...

class FractionFieldElement:
    _numerator: Any
    _denominator: Any
    _is_reduced: bool

    def __init__(self, parent: Any, numerator: Any, denominator: Any = 1, coerce: bool = True, reduce: bool = True) -> None:
        ...

    def _im_gens_(self, codomain: Any, im_gens: Any, base_map: Any = None) -> Any:
        ...

    def reduce(self) -> None:
        ...

    def __copy__(self) -> 'FractionFieldElement':
        ...

    def numerator(self) -> Any:
        ...

    def denominator(self) -> Any:
        ...

    def is_square(self, root: bool = False) -> Union[bool, tuple[bool, Any]]:
        ...

    def nth_root(self, n: int) -> 'FractionFieldElement':
        ...

    def __hash__(self) -> int:
        ...

    def __call__(self, *x: Any, **kwds: Any) -> 'FractionFieldElement':
        ...

    def subs(self, in_dict: dict | None = None, *args: Any, **kwds: Any) -> 'FractionFieldElement':
        ...

    def _is_atomic(self) -> bool:
        ...

    def _repr_(self) -> str:
        ...

    def _latex_(self) -> str:
        ...

    def _magma_init_(self, magma: Any) -> str:
        ...

    def _add_(self, right: 'FractionFieldElement') -> 'FractionFieldElement':
        ...

    def _mul_(self, right: 'FractionFieldElement') -> 'FractionFieldElement':
        ...

    def _div_(self, right: 'FractionFieldElement') -> 'FractionFieldElement':
        ...

    def __int__(self) -> int:
        ...

    def __float__(self) -> float:
        ...

    def __complex__(self) -> complex:
        ...

    def _rational_(self) -> 'FractionFieldElement':
        ...

    def _conversion(self, R: Any) -> 'FractionFieldElement':
        ...

    def __pow__(self, right: int, dummy: Any) -> 'FractionFieldElement':
        ...

    def __neg__(self) -> 'FractionFieldElement':
        ...

    def __abs__(self) -> 'FractionFieldElement':
        ...

    def __invert__(self) -> 'FractionFieldElement':
        ...

    def _richcmp_(self, other: 'FractionFieldElement', op: int) -> bool:
        ...

    def valuation(self, v: Any = None) -> int:
        ...

    def __bool__(self) -> bool:
        ...

    def is_zero(self) -> bool:
        ...

    def is_one(self) -> bool:
        ...

    def _symbolic_(self, ring: Any) -> 'FractionFieldElement':
        ...

    def __reduce__(self) -> tuple:
        ...

    def _evaluate_polynomial(self, pol: Any) -> 'FractionFieldElement':
        ...

    def specialization(self, D: Any = None, phi: Any = None) -> 'FractionFieldElement':
        ...

class FractionFieldElement_1poly_field(FractionFieldElement):
    def __init__(self, parent: Any, numerator: Any, denominator: Any = 1, coerce: bool = True, reduce: bool = True) -> None:
        ...

    def normalize_leading_coefficients(self) -> None:
        ...

    def is_integral(self) -> bool:
        ...

    def support(self) -> Any:
        ...

    def reduce(self) -> None:
        ...

def make_element(parent: Any, numerator: Any, denominator: Any) -> FractionFieldElement:
    ...

def make_element_old(parent: Any, cdict: dict) -> FractionFieldElement:
    ...



# sage_setup: distribution = sagemath-categories
"""
Big O for various types (power series, `p`-adics, etc.)

.. SEEALSO::

    - `asymptotic expansions <../../../asymptotic/index.html>`_
    - `p-adic numbers <../../../padics/index.html>`_
    - `power series <../../../power_series/index.html>`_
    - `polynomials <../../../polynomial_rings/index.html>`_
"""

from sage.misc.lazy_import import lazy_import

lazy_import('sage.rings.padics.factory', ['Qp', 'Zp'])
lazy_import('sage.rings.padics.padic_ZZ_pX_FM_element', 'pAdicZZpXFMElement')
lazy_import('sage.rings.padics.padic_fixed_mod_element', 'pAdicFixedModElement')

from sage.rings.polynomial.polynomial_element import Polynomial

try:
    from .puiseux_series_ring_element import PuiseuxSeries
except ImportError:
    PuiseuxSeries = ()

from sage.rings import (
    multi_power_series_ring_element,
    power_series_ring_element,
)
from sage.rings.integer import Integer
from sage.rings.rational import Rational


def O(*x, **kwds):
    """
    Big O constructor for various types.

    EXAMPLES:

    This is useful for writing power series elements::

        sage: R.<t> = ZZ[['t']]
        sage: (1+t)^10 + O(t^5)
        1 + 10*t + 45*t^2 + 120*t^3 + 210*t^4 + O(t^5)

    A power series ring is created implicitly if a polynomial
    element is passed::

        sage: R.<x> = QQ['x']
        sage: O(x^100)
        O(x^100)
        sage: 1/(1+x+O(x^5))
        1 - x + x^2 - x^3 + x^4 + O(x^5)
        sage: R.<u,v> = QQ[[]]
        sage: 1 + u + v^2 + O(u, v)^5
        1 + u + v^2 + O(u, v)^5

    This is also useful to create `p`-adic numbers::

        sage: O(7^6)                                                                    # needs sage.rings.padics
        O(7^6)
        sage: 1/3 + O(7^6)                                                              # needs sage.rings.padics
        5 + 4*7 + 4*7^2 + 4*7^3 + 4*7^4 + 4*7^5 + O(7^6)

    It behaves well with respect to adding negative powers of `p`::

        sage: a = O(11^-32); a                                                          # needs sage.rings.padics
        O(11^-32)
        sage: a.parent()                                                                # needs sage.rings.padics
        11-adic Field with capped relative precision 20

    There are problems if you add a rational with very negative
    valuation to an `O`-Term::

        sage: 11^-12 + O(11^15)                                                         # needs sage.rings.padics
        11^-12 + O(11^8)

    The reason that this fails is that the constructor doesn't know
    the right precision cap to use. If you cast explicitly or use
    other means of element creation, you can get around this issue::

        sage: # needs sage.rings.padics
        sage: K = Qp(11, 30)
        sage: K(11^-12) + O(11^15)
        11^-12 + O(11^15)
        sage: 11^-12 + K(O(11^15))
        11^-12 + O(11^15)
        sage: K(11^-12, absprec=15)
        11^-12 + O(11^15)
        sage: K(11^-12, 15)
        11^-12 + O(11^15)

    We can also work with `asymptotic expansions`_::

        sage: A.<n> = AsymptoticRing(growth_group='QQ^n * n^QQ * log(n)^QQ',            # needs sage.graphs sage.symbolic
        ....:                        coefficient_ring=QQ); A
        Asymptotic Ring <QQ^n * n^QQ * log(n)^QQ * Signs^n> over Rational Field
        sage: O(n)                                                                      # needs sage.graphs sage.symbolic
        O(n)

    Application with Puiseux series::

        sage: P.<y> = PuiseuxSeriesRing(ZZ)
        sage: y^(1/5) + O(y^(1/3))
        y^(1/5) + O(y^(1/3))
        sage: y^(1/3) + O(y^(1/5))
        O(y^(1/5))


    TESTS::

        sage: var('x, y')                                                               # needs sage.symbolic
        (x, y)
        sage: O(x)                                                                      # needs sage.symbolic
        Traceback (most recent call last):
        ...
        ArithmeticError: O(x) not defined
        sage: O(y)                                                                      # needs sage.symbolic
        Traceback (most recent call last):
        ...
        ArithmeticError: O(y) not defined
        sage: O(x, y)
        Traceback (most recent call last):
        ...
        ArithmeticError: O(x, y) not defined
        sage: O(4, 2)
        Traceback (most recent call last):
        ...
        ArithmeticError: O(4, 2) not defined

    ::

        sage: R.<x> = QQ[]
        sage: O(2*x)
        Traceback (most recent call last):
        ...
        NotImplementedError: completion only currently defined for the maximal ideal (x)
        sage: R.<x> = LazyPowerSeriesRing(QQ)                                           # needs sage.combinat
        sage: O(x^5)                                                                    # needs sage.combinat
        O(x^5)

        sage: # needs sage.rings.padics
        sage: t = O(Zp(5)(2*5^2)); t
        O(5^2)
        sage: t.parent()
        5-adic Ring with capped relative precision 20
        sage: t = O(Qp(5)(2*5^-2)); t
        O(5^-2)
        sage: t.parent()
        5-adic Field with capped relative precision 20
        sage: O(-6)
        Traceback (most recent call last):
        ...
        ArithmeticError: x must be a prime power >= 2
        sage: O(6)
        Traceback (most recent call last):
        ...
        ArithmeticError: x must be prime power
        sage: O(11/2)
        Traceback (most recent call last):
        ...
        ArithmeticError: x must be prime power
        sage: O(Rational(8))
        O(2^3)
    """
    if len(x) > 1:
        if isinstance(x[0], multi_power_series_ring_element.MPowerSeries):
            return multi_power_series_ring_element.MO(x, **kwds)
        else:
            raise ArithmeticError("O(%s) not defined" %
                                  (', '.join(str(e) for e in x),))

    x = x[0]

    if isinstance(x, power_series_ring_element.PowerSeries):
        return x.parent()(0, x.degree(), **kwds)

    if isinstance(x, Polynomial):
        if x.parent().ngens() != 1:
            raise NotImplementedError("completion only currently defined "
                                      "for univariate polynomials")
        if not x.is_monomial():
            raise NotImplementedError("completion only currently defined "
                                      "for the maximal ideal (x)")

    if isinstance(x, (int, Integer, Rational)):
        # p-adic number
        if x <= 0:
            raise ArithmeticError("x must be a prime power >= 2")
        if isinstance(x, (int, Integer)):
            x = Integer(x)
            p, r = x.perfect_power()
        else:
            if x.denominator() == 1:
                p, r = x.numerator().perfect_power()
            elif x.numerator() == 1:
                p, r = x.denominator().perfect_power()
                r = -r
            else:
                raise ArithmeticError("x must be prime power")
        if not p.is_prime():
            raise ArithmeticError("x must be prime power")
        if r >= 0:
            return Zp(p, prec=max(r, 20),
                      type='capped-rel')(0, absprec=r, **kwds)
        else:
            return Qp(p, prec=max(r, 20),
                      type='capped-rel')(0, absprec=r, **kwds)

    if isinstance(x, PuiseuxSeries):
        # note that add_bigoh() of PuiseuxSeries adapts the precision
        # to the ramification index of the input, thus we cannot do
        # zero.add_bigoh() because zero has ramification index 1
        return x.add_bigoh(x.valuation(), **kwds)

    if isinstance(x, (pAdicZZpXFMElement, pAdicFixedModElement)):
        # fixed modulus elements does not keep track of their own precision,
        # we must return zero (that said it is not recommended to use O()
        # in this case, because it does nothing)
        return x.parent().zero()

    try:
        return x.parent().zero().add_bigoh(x.valuation(), **kwds)
    except AttributeError:
        if hasattr(x, 'O'):  # this case is used for AsymptoticRing
            return x.O(**kwds)
        raise ArithmeticError("O(%s) not defined" % (x,))
